from decimal import Decimal

from conseil.query import MetadataQuery, DataQuery
from conseil.api import ConseilException


def not_(predicate: dict):
    res = predicate.copy()
    res['inverse'] = True
    return res


class Attribute(MetadataQuery):
    __query_path__ = 'metadata/{platform_id}/{network_id}/{entity_id}/{attribute_id}'

    def __hash__(self):
        return id(self)

    def __getattr__(self, item):
        return item

    def query(self) -> DataQuery:
        """
        Request specific attribute
        :return: DataQuery
        """
        kwargs = self._extend(attributes={self['attribute_id']: self})
        return DataQuery(self._api, **kwargs)

    def _predicate(self, operation, *args, inverse=False):
        predicate = {
            'field': self['attribute_id'],
            'operation': operation,
            'set': list(args),
            'inverse': inverse
        }

        if args and all(map(lambda x: isinstance(x, Decimal), args)):
            precision = max(map(lambda x: max(0, -x.as_tuple().exponent), args))
            predicate['precision'] = precision

        return predicate

    def _sort_order(self, direction):
        return {
            'field': self['attribute_id'],
            'direction': direction
        }

    def _aggregate(self, function):
        return self._spawn(aggregation={
            'field': self['attribute_id'],
            'function': function
        })

    def in_(self, *args):
        if len(args) == 0:
            return self.is_(None)
        if len(args) == 1:
            return self.is_(args[0])
        return self._predicate('in', *args)

    def notin_(self, *args):
        if len(args) == 0:
            return self.isnot(None)
        if len(args) == 1:
            return self.isnot(args[0])
        return self._predicate('in', *args, inverse=True)

    def is_(self, value):
        if value is None:
            return self._predicate('isnull')
        return self._predicate('eq', value)

    def isnot(self, value):
        if value is None:
            return self._predicate('isnull', inverse=True)
        return self._predicate('eq', value, inverse=True)

    def between(self, first, second):
        return self._predicate('between', first, second)

    def like(self, value):
        return self._predicate('like', value)

    def notlike(self, value):
        return self._predicate('like', value, inverse=True)

    def __lt__(self, other):
        return self._predicate('lt', other)

    def __ge__(self, other):
        return self._predicate('lt', other, inverse=True)

    def __gt__(self, other):
        return self._predicate('gt', other)

    def __le__(self, other):
        return self._predicate('gt', other, inverse=True)

    def __eq__(self, other):
        return self.is_(other)

    def __ne__(self, other):
        return self.isnot(other)

    def startswith(self, value):
        return self._predicate('startsWith', value)

    def endswith(self, value):
        return self._predicate('endsWith', value)

    def asc(self):
        return self._sort_order('asc')

    def desc(self):
        return self._sort_order('desc')

    def sum(self):
        return self._aggregate('sum')

    def count(self):
        return self._aggregate('count')

    def avg(self):
        return self._aggregate('avg')

    def min(self):
        return self._aggregate('min')

    def max(self):
        return self._aggregate('max')

    def label(self, label):
        return self._spawn(label=label)


class Entity(MetadataQuery):
    __child_key__ = 'attribute_id'
    __child_class__ = Attribute
    __query_path__ = 'metadata/{platform_id}/{network_id}/{entity_id}/attributes'

    def query(self, *args) -> DataQuery:
        """__query_path__
        Request an entity or specific fields
        :param args: Array of attributes (of a common entity) or a single entity
        :return: DataQuery
        """
        if all(map(lambda x: isinstance(x, Attribute), args)):
            attributes = {x['attribute_id']: x for x in args}
            if any(map(lambda x: x['entity_id'] != self['entity_id'], args)):
                raise ConseilException('Entity mismatch')
        elif len(args) == 0:
            attributes = dict()
        else:
            raise ConseilException('List of attributes (single entity) or an entity is allowed')
        kwargs = self._extend(attributes=attributes)
        return DataQuery(self._api, **kwargs)


class Network(MetadataQuery):
    __child_key__ = 'entity_id'
    __child_class__ = Entity
    __query_path__ = 'metadata/{platform_id}/{network_id}/entities'

    def query(self, *args) -> DataQuery:
        """
        Request an entity or specific fields
        :param args: Array of attributes (of a common entity) or a single entity
        :return: DataQuery
        """
        if all(map(lambda x: isinstance(x, Attribute), args)):
            attributes = {x['attribute_id']: x for x in args}
            if any(map(lambda x: x['entity_id'] != args[0]['entity_id'], args)):
                raise ConseilException('Mixed entities')
        elif len(args) == 1 and isinstance(args[0], Entity):
            attributes = dict()
        else:
            raise ConseilException('List of attributes (single entity) or an entity is allowed')

        kwargs = self._extend(
            attributes=attributes,
            entity_id=args[0]['entity_id']
        )
        return DataQuery(self._api, **kwargs)


class Platform(MetadataQuery):
    __child_key__ = 'network_id'
    __child_class__ = Network
    __query_path__ = 'metadata/{platform_id}/networks'


class Client(MetadataQuery):
    __child_key__ = 'platform_id'
    __child_class__ = Platform
    __query_path__ = 'metadata/platforms'



import io
import csv
from os.path import basename
from functools import lru_cache
from pprint import pformat

from conseil.api import ConseilApi, ConseilException


def list2csv(data: list):
    fp = io.StringIO()
    writer = csv.DictWriter(fp, fieldnames=data[0].keys())
    writer.writeheader()
    writer.writerows(data)
    return fp.getvalue()


class Query:
    __query_path__ = None

    def __init__(self, api=ConseilApi(), **kwargs):
        self._api = api
        self._kwargs = kwargs

    def _extend(self, **kwargs):
        params = self._kwargs.copy()
        for key, value in kwargs.items():
            if isinstance(params.get(key), list):
                params[key].extend(value)
            else:
                params[key] = value
        return params

    def _spawn(self, **kwargs):
        params = self._extend(**kwargs)
        return self.__class__(self._api, **params)

    @property
    def _query_path(self):
        return self.__query_path__.format(**self._kwargs)

    def __getitem__(self, item):
        return self._kwargs.get(item)

    def __repr__(self):
        if self.__query_path__:
            return f'Path\n{self._query_path}\n\n'


class MetadataQuery(Query):
    __child_key__ = None
    __child_class__ = None

    @lru_cache(maxsize=None)
    def _request(self):
        try:
            if self.__query_path__:
                return self._api.get(self._query_path).json()
        except ConseilException:
            pass
        return list()

    @property
    def _attr_names(self):
        return filter(lambda x: x,
                      map(lambda x: x.get('name', x) if isinstance(x, dict) else x,
                          self._request()))

    def __repr__(self):
        docstring = super(MetadataQuery, self).__repr__()

        attr_names = '\n'.join(map(lambda x: f'.{x}', self._attr_names))
        if attr_names:
            title = basename(self._query_path).capitalize()
            docstring += f'{title}\n{attr_names}\n'

        return docstring

    def __call__(self):
        return self._request()

    def __dir__(self):
        return list(super(MetadataQuery, self).__dir__()) + list(self._attr_names)

    @lru_cache(maxsize=None)
    def __getattr__(self, item):
        if self.__child_class__:
            kwargs = {
                self.__child_key__: item,
                **self._kwargs
            }
            return self.__child_class__(self._api, **kwargs)
        raise ConseilException(item)


class DataQuery(Query):
    __query_path__ = 'data/{platform_id}/{network_id}/{entity_id}'

    def payload(self):
        """
        Resulting Conseil query
        :return: object
        """
        attributes = self['attributes'] or dict()
        having = self['having'] or []
        orders = self['order_by'] or []
        group_by = self['group_by'] or []

        for predicate in having:
            try:
                attributes[predicate['field']]['aggregation']['predicate'] = predicate
            except (KeyError, TypeError):
                raise ConseilException(f'Orphan HAVING predicate on `{predicate["field"]}`')

        aggregation = [x['aggregation'] for x in attributes.values() if x['aggregation']]

        for order in orders:
            try:
                function = attributes[order['field']]['aggregation']['function']
                order['field'] = f'{function}_{order["field"]}'
            except (KeyError, TypeError):
                pass

        fields = list(attributes.keys())

        for attr in group_by:
            if attr['attribute_id'] not in fields:
                fields.append(attr['attribute_id'])

        return {
            'fields': fields,
            'predicates': list(self['predicates'] or []),
            'aggregation': aggregation,
            'orderBy': orders,
            'limit': self['limit'],
            'output': self['output'] or 'json'
        }

    def field_map(self):
        """
        Postprocessing rules: renaming fields, dropping columns
        :return: object
        """
        attributes = self['attributes'] or dict()
        group_by = self['group_by'] or []

        field_map = {
            field: attr['label']
            for field, attr in attributes.items()
            if attr['label']
        }

        for field in group_by:
            field_map[field] = False

        return field_map

    def _postprocess(self, data: list, field_map: dict):
        def process(item):
            return {
                field_map.get(k, k): v
                for k, v in item.items()
                if field_map.get(k) is not False
            }

        return list(map(process, data))

    def __repr__(self):
        docstring = super(DataQuery, self).__repr__()
        docstring += f'Query\n{pformat(self.payload())}\n\n'

        postprocess = '\n'.join(map(lambda x: f'{x[0]} -> {x[1]}', self.field_map().items()))
        if postprocess:
            docstring += f'Postprocess\n{postprocess}\n\n'

        return docstring

    def filter(self, *args):
        """
        Use predicates to filter results (conjunction)
        :param args: array of predicates
        :return: DataQuery
        """
        return self._spawn(predicates=args)

    def filter_by(self, **kwargs):
        """
        Use simple `eq` predicates to filter results (conjunction)
        :param kwargs: pairs <attribute_id>=<value>
        :return: DataQuery
        """
        predicates = [
            {
                'field': k,
                'operation': 'eq',
                'set': [v],
                'inverse': False
            }
            for k, v in kwargs.items()
        ]
        return self._spawn(predicates=predicates)

    def order_by(self, *args):
        """
        Sort results by specified columns
        :param args: one or many sort rules
        :return: DataQuery
        """
        order_by = [x if isinstance(x, dict) else x.asc() for x in args]
        return self._spawn(order_by=order_by)

    def limit(self, limit: int):
        """
        Limit results count
        :param limit: integer
        :return: DataQuery
        """
        return self._spawn(limit=limit)

    def group_by(self, *args):
        """
        If you want to group by these columns but don't want them to be in the result
        :param args: attributes
        :return: DataQuery
        """
        return self._spawn(group_by=args)

    def having(self, *args):
        """
        Filter results by aggregated column
        :param args: array of predicates on aggregated columns
        :return: DataQuery
        """
        return self._spawn(having=args)

    def all(self, output='json'):
        """
        Get all results
        :param output: output format (json/csv), default is JSON
        :return: list (json) or string (csv)
        """
        field_map = self.field_map()
        self._kwargs['output'] = 'json' if field_map else output

        res = self._api.post(path=self._query_path, json=self.payload())
        if field_map:
            data = self._postprocess(res.json(), field_map)
            if output == 'csv':
                data = list2csv(data)
        elif output == 'json':
            data = res.json()
        elif output == 'csv':
            data = res.text
        else:
            raise NotImplementedError(output)

        return data

    def one(self):
        """
        Get single result, fail if there are no or multiple records (json only)
        :return: object
        """
        res = self.all()
        if len(res) == 0:
            raise ConseilException('Not found')
        if len(res) > 1:
            raise ConseilException('Multiple results')
        return res[0]

    def one_or_none(self):
        """
        Get single result or None (do not fail)
        :return: object or None
        """
        try:
            return self.one()
        except ConseilException:
            pass

    def scalar(self):
        """
        Get value of a single attribute (single column, single row)
        :return: scalar
        """
        res = self.one()
        if len(res) != 1:
            raise ConseilException('Multiple keys')
        return next(iter(res.values()))



import requests


class ConseilException(Exception):
    pass


class ConseilApi:

    def __init__(self,
                 api_key='bakingbad',
                 api_host='https://conseil-dev.cryptonomic-infra.tech',
                 api_version=2,
                 timeout=15):
        self._api_key = api_key
        self._api_host = api_host
        self._api_version = api_version
        self._timeout = timeout

    def _request(self, method, path, json=None):
        response = requests.request(
            method=method,
            url=f'{self._api_host}/v{self._api_version}/{path}',
            headers={'apiKey': self._api_key},
            json=json,
            timeout=self._timeout
        )
        if response.status_code != 200:
            raise ConseilException(f'[{response.status_code}]: {response.text}')

        return response

    def get(self, path):
        return self._request(method='GET', path=path)

    def post(self, path, json):
        return self._request(method='POST', path=path, json=json)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['conseil']

package_data = \
{'': ['*']}

install_requires = \
['requests']

setup_kwargs = {
    'name': 'conseil',
    'version': '0.1.1',
    'description': 'Python toolkit for Conseil blockchain indexer',
    'long_description': '# ConseilPy\n[![PyPI version](https://badge.fury.io/py/conseil.svg?)](https://badge.fury.io/py/conseil)\n[![Build Status](https://travis-ci.org/baking-bad/conseilpy.svg?branch=master)](https://travis-ci.org/baking-bad/conseilpy)\n[![Made With](https://img.shields.io/badge/made%20with-python-blue.svg?)](https://www.python.org)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\nPython toolkit for [Conseil](https://cryptonomic.github.io/Conseil) blockchain indexer\n\n## Installation\n\nPython 3.6+ required\n\n```bash\n$ pip install conseil\n```\n\n## Usage\n\nConseilPy is a lot like Sqlalchemy, so if you\'re familiar with it, you can easily cook queries.\n\n![It\'s time to cook](https://memegenerator.net/img/instances/48258954.jpg)\n\n### Quickstart\nGet top 5 delegators by balance\n```python\nfrom conseil import conseil\n\nAccount = conseil.tezos.alphanet.accounts\nAccount.query(Account.acccount_id, Account.balance) \\\n    .filter(Account.script.is_(None), \n            Account.account_id.startswith(\'KT1\')) \\\n    .order_by(Account.balance.desc()) \\\n    .limit(5) \\\n    .all()\n```\n\nSee more [examples](https://github.com/baking-bad/conseilpy/tree/master/examples)\n\n### Client initialization\nIf using a default conseil client is not an option you can instantiate it yourself:\n```python\nfrom conseil.api import ConseilApi\nfrom conseil.core import Client\n\nconseil = Client(ConseilApi(\n    api_key=\'<API_KEY>\',\n    api_host=\'<API_HOST>\',\n    api_version=2\n))\n```\n\n### Exploring database schema\nConseil metadata has the following tree structure:  \nplatform / network / entity / attribute / value\n\nSo you can simply access any node by name:\n```python\n>>> from conseil import conseil\n>>> print(conseil.tezos.alphanet.operations.kind.transaction)\ntransaction\n```\n\nAutocompletion `Shift + Tab` and docstrings are available in Jupyter:\n```python\n>>> conseil\nPath\nmetadata/platforms\n\nPlatforms\n.tezos\n\n>>> conseil.tezos.alphanet\nPath\nmetadata/tezos/alphanet/entities\n\nEntities\n.accounts\n.balance_updates\n.ballots\n.blocks\n.delegates\n.fees\n.operation_groups\n.operations\n.proposals\n.rolls\n```\n\nAlternatively you can check full [SQL schema](https://github.com/Cryptonomic/Conseil/blob/master/doc/conseil.sql)\n\n### Selecting fields\nConseil doesn\'t support joins at the moment so you can request attributes for a single entity only. \n\n```python\nfrom conseil import conseil\n\nc = conseil.tezos.alphanet\n\n# select all fields\nc.query(c.accounts)\nc.accounts.query()\n\n# select specific fields\nc.query(c.accounts.balance, c.accounts.account_id)\nc.accounts.query(c.accounts.balance, c.accounts.account_id)\n\n# select single field\nc.accounts.balance.query()\n```\n\n### Filtering results\nConseil receives a conjunction of predicates, which can be inverted by one, but not together.\nPredicate syntax is similar to Sqlalchemy, but has less operations.\n\n```python\nfrom conseil import conseil\nfrom conseil.core import not_\n\nAccount = conseil.tezos.alphanet.accounts\nAccount.query() \\\n    .filter(not_(Account.account_id.startswith(\'tz\')),\n            Account.script.is_(None),\n            Account.balance > 0)\n```\n\nHere is a full list of supported operations:\n\n| Conseil operation | Filter             | Inversed                |\n| ----------------- | ------------------ | ----------------------- |\n| in                | `x.in_(a, b, ...)` | `x.notin_(a, b, ...)`   |\n| between           | `x.between(a, b)`  | `not_(x.between(a, b))` |\n| like              | `x.like(a)`        | `x.notlike(a)`          |\n| lt                | `x < a`            | `x >= a`                |\n| gt                | `x > a`            | `x <= a`                |\n| eq                | `x == a`           | `x != a`                |\n| startsWith        | `x.startswith(a)`  | `not_(x.startsWith(a))` |\n| endsWith          | `x.endswith(a)`    | `not_(x.endswith(a))`   |\n| isnull            | `x.is_(None)`      | `x.isnot(None)`         |\n\nYou can also use `filter_by` for simple queries:\n\n```python\nfrom conseil import conseil\n\nconseil.tezos.alphanet.accounts.query() \\\n    .filter_by(account_id=\'tzkt\')\n```\n\n### Data aggregation\n\nThis is an important concept to understand. In Conseil you specify which columns will be aggregated and the rest of them are used in `GROUP BY` clause. Here is an example:\n\n```python\nfrom conseil import conseil\n\nBlock = conseil.tezos.alphanet.blocks\nBlock.query(Block.baker, Block.level.count())  \n# will be compiled to SELECT baker, COUNT(level) FROM blocks GROUP BY baker\n```\n\nAdditionally, you can specify `HAVING` predicates if you want to filter results by aggregated column:\n\n```python\nfrom conseil import conseil\n\nBlock = conseil.tezos.alphanet.blocks\nBlock.query(Block.baker, Block.level.count()) \\\n    .having(Block.level > 1)  # or Block.level.count(), no difference\n```\n\nHere is the list of supported aggregation functions:\n\n* `count`\n* `sum`\n* `avg`\n* `min`\n* `max`\n\nIf you want to group by some fields but not include them in the result use `group_by` method:\n\n```python\nfrom conseil import conseil\n\nBlock = conseil.tezos.alphanet.blocks\nBlock.query(Block.level.count()) \\\n\t.group_by(Block.baker)\n```\n\n### Sorting and limiting results\n\nThis is similar to Sqlalchemy as well, you can specify one or multiple sort columns with optional descending modifier.\n\n```python\nfrom conseil import conseil\n\nAccount = conseil.tezos.alphanet.accounts\nAccount.query() \\\n    .order_by(Account.balance.desc(), Account.account_id) \\\n    .limit(20)\n```\n\n### Query preview\n\nSo you have cooked a simple query and want to see the resulting Conseil request body.\n\n```python\nfrom conseil import conseil\n\nAccount = conseil.tezos.alphanet.accounts\nquery = Account.query() \\\n    .order_by(Account.balance.desc()) \\\n    .limit(1)\n```\n\nThen you can simply:\n\n```python\n>>> query\nPath\ndata/tezos/alphanet/accounts\n\nQuery\n{"aggregation": [],\n "fields": [],\n "limit": 1,\n "orderBy": [{"direction": "desc", "field": "balance"}],\n "output": "json",\n "predicates": []}\n```\n\n### Execution\n\nIt\'s time to submit our query and get some data.\n\n```python\nfrom conseil import conseil\n\nAccount = conseil.tezos.alphanet.accounts\n```\n\n#### Return multiple rows\n\n```python\nquery = Account.query()\n\nquery.all()  # will return List[dict] (default output type)\nquery.all(output=\'csv\')  # will return string (csv)\n```\n\n#### Return single row\n\n```python\nquery = Account.query() \\\n\t.filter_by(account_id=\'tzkt\')\n\nquery.one()  # will fail if there is no account with such id or there are many\nquery.one_or_none()  # will handle the exception and return None\n```\n\n#### Return scalar\n\n```python\nquery = Account.balance.query() \\\n\t.order_by(Account.balance.desc()) \\\n    .limit(1)\n\nquery.scalar()  # will return single numeric value\n```\n\n### Precision\nConseil allows to specify numeric column precision. In order to use this functionality use `decimal` type. For example:\n\n```python\nfrom conseil import conseil\nfrom decimal import Decimal\n\nAccount = conseil.tezos.alphanet.accounts\n\nAccount.query(Account.balance) \\\n    .filter(Account.balance > Decimal(\'0.1\'), \n            Account.balance < Decimal(\'0.01\'))  # precision will be 2 (max)\n```\n\n### Renaming fields\nYou can change names of requested fields in the resulting json/csv:\n\n```python\nfrom conseil import conseil\n\nAccount = conseil.tezos.alphanet.accounts\n\nAccount.query(Account.account_id.label(\'address\'))\n```',
    'author': 'Michael Zaikin',
    'author_email': 'mz@baking-bad.org',
    'url': 'https://github.com/baking-bad/conseilpy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
